@@ -29,11 +29,12 @@ from __future__ import annotations
 
 import re
 from collections import OrderedDict, defaultdict
-from collections.abc import Iterable
 from dataclasses import dataclass
 from datetime import date
-from typing import TYPE_CHECKING
+from itertools import chain
+from typing import TYPE_CHECKING, Any
 
+from deprecated import deprecated
 from jinja2 import (
     BaseLoader,
     ChoiceLoader,
@@ -42,13 +43,14 @@ from jinja2 import (
     Template,
 )
 
-from commitizen.cz.base import ChangelogReleaseHook
 from commitizen.exceptions import InvalidConfigurationError, NoCommitsFoundError
-from commitizen.git import GitCommit, GitTag
 from commitizen.tags import TagRules
 
 if TYPE_CHECKING:
-    from commitizen.cz.base import MessageBuilderHook
+    from collections.abc import Generator, Iterable, Mapping, MutableMapping, Sequence
+
+    from commitizen.cz.base import ChangelogReleaseHook, MessageBuilderHook
+    from commitizen.git import GitCommit, GitTag
 
 
 @dataclass
@@ -63,13 +65,26 @@ class Metadata:
     latest_version_position: int | None = None
     latest_version_tag: str | None = None
 
-    def __post_init__(self):
+    def __post_init__(self) -> None:
         if self.latest_version and not self.latest_version_tag:
             # Test syntactic sugar
             # latest version tag is optional if same as latest version
             self.latest_version_tag = self.latest_version
 
 
+@dataclass
+class IncrementalMergeInfo:
+    """
+    Information regarding the last non-pre-release, parsed from the changelog.
+
+    Required to merge pre-releases on bump.
+    Separate from Metadata to not mess with the interface.
+    """
+
+    name: str | None = None
+    index: int | None = None
+
+
 def get_commit_tag(commit: GitCommit, tags: list[GitTag]) -> GitTag | None:
     return next((tag for tag in tags if tag.rev == commit.rev), None)
 
@@ -84,37 +99,39 @@ def generate_tree_from_commits(
     changelog_message_builder_hook: MessageBuilderHook | None = None,
     changelog_release_hook: ChangelogReleaseHook | None = None,
     rules: TagRules | None = None,
-) -> Iterable[dict]:
+    during_version_bump: bool = False,
+) -> Generator[dict[str, Any], None, None]:
     pat = re.compile(changelog_pattern)
     map_pat = re.compile(commit_parser, re.MULTILINE)
     body_map_pat = re.compile(commit_parser, re.MULTILINE | re.DOTALL)
-    current_tag: GitTag | None = None
     rules = rules or TagRules()
 
     # Check if the latest commit is not tagged
-    if commits:
-        latest_commit = commits[0]
-        current_tag = get_commit_tag(latest_commit, tags)
+    if during_version_bump and rules.merge_prereleases:
+        current_tag = None
+    else:
+        current_tag = get_commit_tag(commits[0], tags) if commits else None
+    current_tag_name = unreleased_version or "Unreleased"
+    current_tag_date = (
+        date.today().isoformat() if unreleased_version is not None else ""
+    )
 
-    current_tag_name: str = unreleased_version or "Unreleased"
-    current_tag_date: str = ""
-    if unreleased_version is not None:
-        current_tag_date = date.today().isoformat()
-    if current_tag is not None and current_tag.name:
-        current_tag_name = current_tag.name
-        current_tag_date = current_tag.date
+    used_tags: set[GitTag] = set()
+    if current_tag:
+        used_tags.add(current_tag)
+        if current_tag.name:
+            current_tag_name = current_tag.name
+            current_tag_date = current_tag.date
 
+    commit_tag: GitTag | None = None
     changes: dict = defaultdict(list)
-    used_tags: list = [current_tag]
     for commit in commits:
-        commit_tag = get_commit_tag(commit, tags)
-
         if (
-            commit_tag
+            (commit_tag := get_commit_tag(commit, tags))
             and commit_tag not in used_tags
             and rules.include_in_changelog(commit_tag)
         ):
-            used_tags.append(commit_tag)
+            used_tags.add(commit_tag)
             release = {
                 "version": current_tag_name,
                 "date": current_tag_date,
@@ -127,24 +144,15 @@ def generate_tree_from_commits(
             current_tag_date = commit_tag.date
             changes = defaultdict(list)
 
-        matches = pat.match(commit.message)
-        if not matches:
+        if not pat.match(commit.message):
             continue
 
-        # Process subject from commit message
-        if message := map_pat.match(commit.message):
-            process_commit_message(
-                changelog_message_builder_hook,
-                message,
-                commit,
-                changes,
-                change_type_map,
-            )
-
-        # Process body from commit message
-        body_parts = commit.body.split("\n\n")
-        for body_part in body_parts:
-            if message := body_map_pat.match(body_part):
+        # Process subject and body from commit message
+        for message in chain(
+            [map_pat.match(commit.message)],
+            (body_map_pat.match(block) for block in commit.body.split("\n\n")),
+        ):
+            if message:
                 process_commit_message(
                     changelog_message_builder_hook,
                     message,
@@ -167,10 +175,10 @@ def process_commit_message(
     hook: MessageBuilderHook | None,
     parsed: re.Match[str],
     commit: GitCommit,
-    changes: dict[str | None, list],
-    change_type_map: dict[str, str] | None = None,
-):
-    message: dict = {
+    ref_changes: MutableMapping[str | None, list],
+    change_type_map: Mapping[str, str] | None = None,
+) -> None:
+    message: dict[str, Any] = {
         "sha1": commit.rev,
         "parents": commit.parents,
         "author": commit.author,
@@ -178,33 +186,39 @@ def process_commit_message(
         **parsed.groupdict(),
     }
 
-    if processed := hook(message, commit) if hook else message:
-        messages = [processed] if isinstance(processed, dict) else processed
-        for msg in messages:
-            change_type = msg.pop("change_type", None)
-            if change_type_map:
-                change_type = change_type_map.get(change_type, change_type)
-            changes[change_type].append(msg)
+    processed_msg = hook(message, commit) if hook else message
+    if not processed_msg:
+        return
+
+    messages = [processed_msg] if isinstance(processed_msg, dict) else processed_msg
+    for msg in messages:
+        change_type = msg.pop("change_type", None)
+        if change_type_map:
+            change_type = change_type_map.get(change_type, change_type)
+        ref_changes[change_type].append(msg)
 
 
-def order_changelog_tree(tree: Iterable, change_type_order: list[str]) -> Iterable:
+def generate_ordered_changelog_tree(
+    tree: Iterable[Mapping[str, Any]], change_type_order: list[str]
+) -> Generator[dict[str, Any], None, None]:
     if len(set(change_type_order)) != len(change_type_order):
         raise InvalidConfigurationError(
-            f"Change types contain duplicates types ({change_type_order})"
+            f"Change types contain duplicated types ({change_type_order})"
         )
 
-    sorted_tree = []
     for entry in tree:
-        ordered_change_types = change_type_order + sorted(
-            set(entry["changes"].keys()) - set(change_type_order)
-        )
-        changes = [
-            (ct, entry["changes"][ct])
-            for ct in ordered_change_types
-            if ct in entry["changes"]
-        ]
-        sorted_tree.append({**entry, **{"changes": OrderedDict(changes)}})
-    return sorted_tree
+        yield {
+            **entry,
+            "changes": _calculate_sorted_changes(change_type_order, entry["changes"]),
+        }
+
+
+def _calculate_sorted_changes(
+    change_type_order: list[str], changes: Mapping[str, Any]
+) -> OrderedDict[str, Any]:
+    remaining_change_types = set(changes.keys()) - set(change_type_order)
+    sorted_change_types = change_type_order + sorted(remaining_change_types)
+    return OrderedDict((ct, changes[ct]) for ct in sorted_change_types if ct in changes)
 
 
 def get_changelog_template(loader: BaseLoader, template: str) -> Template:
@@ -222,7 +236,7 @@ def render_changelog(
     tree: Iterable,
     loader: BaseLoader,
     template: str,
-    **kwargs,
+    **kwargs: Any,
 ) -> str:
     jinja_template = get_changelog_template(loader, template)
     changelog: str = jinja_template.render(tree=tree, **kwargs)
@@ -248,6 +262,7 @@ def incremental_build(
     unreleased_start = metadata.unreleased_start
     unreleased_end = metadata.unreleased_end
     latest_version_position = metadata.latest_version_position
+
     skip = False
     output_lines: list[str] = []
     for index, line in enumerate(lines):
@@ -257,9 +272,7 @@ def incremental_build(
             skip = False
             if (
                 latest_version_position is None
-                or isinstance(latest_version_position, int)
-                and isinstance(unreleased_end, int)
-                and latest_version_position > unreleased_end
+                or latest_version_position > unreleased_end
             ):
                 continue
 
@@ -268,18 +281,34 @@ def incremental_build(
 
         if index == latest_version_position:
             output_lines.extend([new_content, "\n"])
-
         output_lines.append(line)
-    if not isinstance(latest_version_position, int):
-        if output_lines and output_lines[-1].strip():
-            # Ensure at least one blank line between existing and new content.
-            output_lines.append("\n")
-        output_lines.append(new_content)
+
+    if latest_version_position is not None:
+        return output_lines
+
+    if output_lines and output_lines[-1].strip():
+        # Ensure at least one blank line between existing and new content.
+        output_lines.append("\n")
+    output_lines.append(new_content)
     return output_lines
 
 
+def get_next_tag_name_after_version(tags: Iterable[GitTag], version: str) -> str | None:
+    it = iter(tag.name for tag in tags)
+    for name in it:
+        if name == version:
+            return next(it, None)
+
+    raise NoCommitsFoundError(f"Could not find a valid revision range. {version=}")
+
+
+@deprecated(
+    reason="This function is unused and will be removed in v5",
+    version="5.0.0",
+    category=DeprecationWarning,
+)
 def get_smart_tag_range(
-    tags: list[GitTag], newest: str, oldest: str | None = None
+    tags: Sequence[GitTag], newest: str, oldest: str | None = None
 ) -> list[GitTag]:
     """Smart because it finds the N+1 tag.
 
@@ -305,49 +334,36 @@ def get_smart_tag_range(
 
 
 def get_oldest_and_newest_rev(
-    tags: list[GitTag],
+    tags: Iterable[GitTag],
     version: str,
     rules: TagRules,
-) -> tuple[str | None, str | None]:
+) -> tuple[str | None, str]:
     """Find the tags for the given version.
 
     `version` may come in different formats:
     - `0.1.0..0.4.0`: as a range
     - `0.3.0`: as a single version
     """
-    oldest: str | None = None
-    newest: str | None = None
-    try:
-        oldest, newest = version.split("..")
-    except ValueError:
-        newest = version
-    if not (newest_tag := rules.find_tag_for(tags, newest)):
+    oldest_version, sep, newest_version = version.partition("..")
+    if not sep:
+        newest_version = version
+        oldest_version = ""
+
+    def get_tag_name(v: str) -> str:
+        if tag := rules.find_tag_for(tags, v):
+            return tag.name
         raise NoCommitsFoundError("Could not find a valid revision range.")
 
-    oldest_tag = None
-    oldest_tag_name = None
-    if oldest:
-        if not (oldest_tag := rules.find_tag_for(tags, oldest)):
-            raise NoCommitsFoundError("Could not find a valid revision range.")
-        oldest_tag_name = oldest_tag.name
+    newest_tag_name = get_tag_name(newest_version)
+    oldest_tag_name = get_tag_name(oldest_version) if oldest_version else None
 
-    tags_range = get_smart_tag_range(
-        tags, newest=newest_tag.name, oldest=oldest_tag_name
+    oldest_rev = get_next_tag_name_after_version(
+        tags, oldest_tag_name or newest_tag_name
     )
-    if not tags_range:
-        raise NoCommitsFoundError("Could not find a valid revision range.")
 
-    oldest_rev: str | None = tags_range[-1].name
-    newest_rev = newest_tag.name
-
-    # check if it's the first tag created
-    # and it's also being requested as part of the range
-    if oldest_rev == tags[-1].name and oldest_rev == oldest_tag_name:
-        return None, newest_rev
-
-    # when they are the same, and it's also the
-    # first tag created
-    if oldest_rev == newest_rev:
-        return None, newest_rev
-
-    return oldest_rev, newest_rev
+    # Return None for oldest_rev if:
+    # 1. The oldest tag is the last tag in the list and matches the requested oldest tag
+    # 2. The oldest and the newest tag are the same
+    if oldest_rev == newest_tag_name:
+        return None, newest_tag_name
+    return oldest_rev, newest_tag_name
